#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <regex.h>

#include "config.h"
#include "imapfilter.h"
#include "account.h"
#include "struct.h"


extern unsigned int flags;

account_t *accounts = NULL;	/* First node of accounts linked list. */
mboxgrp_t *mboxgrps = NULL;	/* First node of mailbox-groups tree. */

account_t *cur_acct = NULL;	/* Current account. */


void init_account(account_t * node);

void init_mboxgrp(mboxgrp_t * node);
void process_mboxgrp(mboxgrp_t * node, char *mboxs);

void init_mbox(mbox_t * node);
mbox_t *set_mbox(char *name);

int string_decode(char *str);


/*
 * Set new account's variables to safe values.
 */
void
init_account(account_t * node)
{
	node->next = NULL;
	node->key[0] = node->server[0] = '\0';
	node->username[0] = node->password[0] = '\0';
	node->passwdattr = PASSWORD_NONE;
	node->port = 143;
	node->ssl = SSL_DISABLED;
	node->mboxes = NULL;
}


/*
 * A new account entry was declared. Create it and set it's variables
 * accordingly.
 */
int
set_account(char *line, regmatch_t * m)
{
	int n;
	char p[6];
	account_t *node;

	node = (account_t *) xmalloc(sizeof(account_t));
	node->password = (char *)smalloc(PASSWORD_LEN);

	init_account(node);

	strncat(node->key, line + m[1].rm_so,
	    min(m[1].rm_eo - m[1].rm_so, KEY_LEN - 1));

#ifdef DEBUG
	fprintf(stderr, "debug: ACCOUNT: '%s'\n", node->key);
#endif

	if (m[3].rm_so != -1) {
		strncat(node->username, line + m[3].rm_so,
		    min(m[3].rm_eo - m[3].rm_so, USERNAME_LEN - 1));
	} else {
		strncat(node->username, line + m[5].rm_so,
		    min(m[5].rm_eo - m[5].rm_so, USERNAME_LEN - 1));
	}
	if (strchr(node->username, '%'))
		if (string_decode(node->username))
			return ERROR_CONFIG_PARSE;

#ifdef DEBUG
	fprintf(stderr, "debug: USERNAME: '%s'\n", node->username);
#endif

	if (m[4].rm_so != -1) {
		strncat(node->password, line + m[4].rm_so,
		    min(m[4].rm_eo - m[4].rm_so, PASSWORD_LEN - 1));
		if (strchr(node->password, '%'))
			if (string_decode(node->password))
				return ERROR_CONFIG_PARSE;
		node->passwdattr = PASSWORD_PLAIN;
	} else
		flags |= FLAG_BLANK_PASSWORD;

#ifdef DEBUG
	fprintf(stderr, "debug: PASSWORD: '%s'\n", node->password);
#endif

	strncat(node->server, line + m[6].rm_so,
	    min(m[6].rm_eo - m[6].rm_so, SERVER_LEN - 1));

#ifdef DEBUG
	fprintf(stderr, "debug: SERVER: '%s'\n", node->server);
#endif

	if (m[7].rm_so != -1) {
		n = min(m[7].rm_eo - m[7].rm_so - 1, 5);
		xstrncpy(p, line + m[7].rm_so + 1, n);
		p[n] = '\0';
		node->port = strtoul(p, NULL, 10);
#ifdef DEBUG
		fprintf(stderr, "debug: PORT: %d\n", node->port);
#endif
	}
	if (m[8].rm_so != -1) {
		if (m[7].rm_so == -1)
			node->port = 993;

		if (strcasestr(line + m[8].rm_so, "SSL2"))
			node->ssl = SSL_SSL_V2;
		else if (strcasestr(line + m[8].rm_so, "SSL3"))
			node->ssl = SSL_SSL_V3;
		else
			node->ssl = SSL_TLS_V1;
	}
	APPEND_LINKED_LIST(accounts, node, account);
	cur_acct = node;

	return 0;
}


#ifdef ENCRYPTED_PASSWORDS
/*
 * Find accounts without a password (candicates for password encryption).
 */
char *
find_password(char *user, char *serv)
{
	account_t *a;

	for (a = accounts; a != NULL; a = a->next)
		if (a->passwdattr == PASSWORD_NONE &&
		    !strcmp(a->server, serv) && !strcmp(a->username, user)) {
			a->passwdattr = PASSWORD_ENCRYPTED;
			return a->password;
		}
	return NULL;
}
#endif


/*
 * Set new mailbox-group's variables to safe values.
 */
void
init_mboxgrp(mboxgrp_t * node)
{
	node->left = node->right = NULL;
	node->key[0] = '\0';
	node->mboxes[0] = NULL;
}


/*
 * A new mailbox-group entry was declared. Create it and set it's variables
 * accordingly.
 */
int
set_mboxgrp(char *line, regmatch_t * m)
{
	mboxgrp_t *node;
	char mboxs[LINE_MAX];

	mboxs[0] = '\0';

	if (accounts == NULL)
		return ERROR_CONFIG_PARSE;

	node = (mboxgrp_t *) xmalloc(sizeof(mboxgrp_t));

	init_mboxgrp(node);

	strncat(node->key, line + m[1].rm_so,
	    min(m[1].rm_eo - m[1].rm_so, KEY_LEN - 1));

#ifdef DEBUG
	fprintf(stderr, "debug: FOLDER: '%s'\n", node->key);
#endif

	strncat(mboxs, line + m[2].rm_so,
	    min(m[2].rm_eo - m[2].rm_so, LINE_MAX - 1));

	process_mboxgrp(node, mboxs);

	INSERT_TREE(mboxgrps, node, mboxgrp);

	return 0;
}


/*
 * Calls set_mbox() in order to create mailboxes that are part of
 * the mailbox-group.
 */
void
process_mboxgrp(mboxgrp_t * node, char *mboxs)
{
	unsigned int i;
	char *tok;

	i = 0;

	tok = strtok_r(mboxs, ",", &mboxs);
	while (i < MBOXGRP_MBOXES_MAX - 1 && tok != NULL) {
		node->mboxes[i] = (mbox_t *) set_mbox(tok);
		node->mboxes[++i] = NULL;

		tok = strtok_r(NULL, ",", &mboxs);
	}
}


/*
 * Set new mailbox's variables to safe values.
 */
void
init_mbox(mbox_t * node)
{
	node->next = NULL;
	node->name[0] = '\0';
	node->filters[0] = NULL;
}


/*
 * A new mailbox was declared, create it and set it's variables accordingly.
 */
mbox_t *
set_mbox(char *name)
{
	char s[MBOX_NAME_LEN];
	mbox_t *node, *m;

	*s = '\0';

	if (*name == '"' && *(name + strlen(name) - 1) == '"')
		strncat(s, name + 1, min(strlen(name) - 2, MBOX_NAME_LEN - 1));
	else
		strncat(s, name, min(strlen(name), MBOX_NAME_LEN - 1));

	for (m = cur_acct->mboxes; m != NULL; m = m->next)
		if (!strcmp(m->name, s))
			return m;

	node = (mbox_t *) xmalloc(sizeof(mbox_t));

	init_mbox(node);

	strncat(node->name, s, MBOX_NAME_LEN - 1);

#ifdef DEBUG
	fprintf(stderr, "debug: MBOX: '%s'\n", node->name);
#endif

	APPEND_LINKED_LIST(cur_acct->mboxes, node, mbox);

	return node;
}


/*
 * Decode the character triplet, consisting of the character '%'
 * followed by two hexadecimal digits to its corresponding ASCII
 * character (described in Section 2.2 of RFC 1738).
 */
int
string_decode(char *str)
{
	char *c;
	char hex[3];

	c = str;

	while (*c != '\0') {
		if (*c == '%') {
			if (!isxdigit((unsigned char)(*(c + 1))) ||
			    !isxdigit((unsigned char)(*(c + 2))))
				return ERROR_CONFIG_PARSE;

			xstrncpy(hex, ++c, 2);
			hex[2] = '\0';

			if (!isprint((unsigned char)(*str = (char)strtoul(hex,
					NULL, 16))))
				return ERROR_CONFIG_PARSE;

			str++;
			c += 2;
		} else
			*(str++) = *(c++);
	}
	*str = '\0';

	return 0;
}


/*
 * Convert the names of personal mailboxes using the namespace specified
 * by the mail server.
 */
char *
apply_namespace(char *mbox, char *prefix, char delim)
{
	static char m[MBOX_NAME_LEN];
	char *c;

	if ((prefix[0] == '\0' && delim == '\0') ||
	    (prefix[0] == '\0' && delim == '/') ||
	    !strcasecmp(mbox, "INBOX"))
		return mbox;

	m[0] = '\0';
	strncat(m, prefix, MBOX_NAME_LEN - 1);
	strncat(m, mbox, MBOX_NAME_LEN - strlen(m) - 1);

	c = m;
	while ((c = strchr(c, '/')))
		*(c++) = delim;

#ifdef DEBUG
	fprintf(stderr, "debug: MAILBOX: '%s'\n", m);
#endif

	return m;
}
